#include <cmath>

#ifdef __APPLE__
#include <GLUT/glut.h> // Apple's version of GLUT is here
#else
#include <GL/glut.h> // but the standard place for it is here
#endif

#include "drawing_primitives.h"

// given an RGB color, and a specification of GL_FRONT, GL_BACK, or GL_FRONT_AND_BACK,
// come up with a fairly generic material of the right colour.
static void set_generic_material(float r, float g, float b, GLenum face)
{
   GLfloat ambient[4], diffuse[4], specular[4];
   ambient[0]=0.1*r+0.03; ambient[1]=0.1*g+0.03; ambient[2]=0.1*b+0.03; ambient[3]=1;
   diffuse[0]=0.7*r;      diffuse[1]=0.7*g;      diffuse[2]=0.7*b;      diffuse[3]=1;
   specular[0]=0.1*r+0.1; specular[1]=0.1*g+0.1; specular[2]=0.1*b+0.1; specular[3]=1;
   glMaterialfv(face, GL_AMBIENT, ambient);
   glMaterialfv(face, GL_DIFFUSE, diffuse);
   glMaterialfv(face, GL_SPECULAR, specular);
   glMaterialf(face, GL_SHININESS, 32);
}

// draw a paralellogram, subdivided into n*n subpanels, starting at (x,y,z) and extending in
// to (x,y,z)+(u0,v0,w0) along one edge and to (x,y,z)+(u1,v1,w1) along another, but project
// all points to the unit sphere centered at the origin.
static void draw_sphere_panel(float x, float y, float z,
                              float u0, float v0, float w0,
                              float u1, float v1, float w1,
                              int n)
{
   for(int j=0; j<n; ++j){
      float dj0=j/(float)n, dj1=(j+1)/(float)n;
      glBegin(GL_TRIANGLE_STRIP);
      for(int i=0; i<=n; ++i){
         float di=i/(float)n;
         float a=x+di*u0+dj0*u1, b=y+di*v0+dj0*v1, c=z+di*w0+dj0*w1;
         float m=std::sqrt(a*a+b*b+c*c);
         a/=m; b/=m; c/=m;
         glNormal3f(a,b,c);
         glVertex3f(a,b,c);
         a=x+di*u0+dj1*u1; b=y+di*v0+dj1*v1; c=z+di*w0+dj1*w1;
         m=std::sqrt(a*a+b*b+c*c);
         a/=m; b/=m; c/=m;
         glNormal3f(a,b,c);
         glVertex3f(a,b,c);
      }
      glEnd();
   }
}

void draw_sphere(float r, float g, float b)
{
   set_generic_material(r, g, b, GL_FRONT);
   set_generic_material(1-r, 1-g, 1-b, GL_BACK); // show up problems with orientation

   // glutSolidSphere(1, , ) would be an alternative, but its tesselation looks really ugly.
   // Instead we'll tesselate a cube and project it onto the surface of the sphere
   const int n=5;

   draw_sphere_panel(-1, -1, -1,
                     0, 2, 0,
                     0, 0, 2, n);
   draw_sphere_panel(1, -1, -1,
                     0, 0, 2,
                     0, 2, 0, n);

   draw_sphere_panel(-1, -1, -1,
                     0, 0, 2,
                     2, 0, 0, n);
   draw_sphere_panel(-1, 1, -1,
                     2, 0, 0,
                     0, 0, 2, n);

   draw_sphere_panel(-1, -1, -1,
                     2, 0, 0,
                     0, 2, 0, n);
   draw_sphere_panel(-1, -1, 1,
                     0, 2, 0,
                     2, 0, 0, n);
}

void draw_cube(float r, float g, float b)
{
   set_generic_material(r, g, b, GL_FRONT);
   set_generic_material(1-r, 1-g, 1-b, GL_BACK); // show up problems with orientation

   glBegin(GL_QUADS);

   glNormal3f(-1, 0, 0);
   glVertex3f(-1, -1, -1);
   glVertex3f(-1, -1, 1);
   glVertex3f(-1, 1, 1);
   glVertex3f(-1, 1, -1);

   glNormal3f(1, 0, 0);
   glVertex3f(1, -1, -1);
   glVertex3f(1, 1, -1);
   glVertex3f(1, 1, 1);
   glVertex3f(1, -1, 1);

   glNormal3f(0, -1, 0);
   glVertex3f(-1, -1, -1);
   glVertex3f(1, -1, -1);
   glVertex3f(1, -1, 1);
   glVertex3f(-1, -1, 1);

   glNormal3f(0, 1, 0);
   glVertex3f(-1, 1, -1);
   glVertex3f(-1, 1, 1);
   glVertex3f(1, 1, 1);
   glVertex3f(1, 1, -1);

   glNormal3f(0, 0, -1);
   glVertex3f(-1, -1, -1);
   glVertex3f(-1, 1, -1);
   glVertex3f(1, 1, -1);
   glVertex3f(1, -1, -1);

   glNormal3f(0, 0, 1);
   glVertex3f(-1, -1, 1);
   glVertex3f(1, -1, 1);
   glVertex3f(1, 1, 1);
   glVertex3f(-1, 1, 1);

   glEnd();
}

void draw_cone(float r, float g, float b)
{
   set_generic_material(r, g, b, GL_FRONT);
   set_generic_material(1-r, 1-g, 1-b, GL_BACK); // show up problems with orientation

   const int n=16;

   // do the pointy bit of the cone
   // this would more naturally be done with a triangle fan, but that doesn't work with how we want to do normals
   glBegin(GL_QUAD_STRIP);
   for(int i=0; i<=n; ++i){
      float theta=2*M_PI*(i/(float)n);
      float c=std::cos(theta), s=std::sin(theta);
      glNormal3f(0.707106781186548*c, 0.707106781186548*s, 0.707106781186548);
      glVertex3f(c, s, 0);
      glVertex3f(0, 0, 1);
   }
   glEnd();

   // do the flat bit of the cone
   glBegin(GL_TRIANGLE_FAN);
   glNormal3f(0, 0, -1);
   glVertex3f(1, 0, 0);
   for(int i=1; i<=n; ++i){
      float theta=2*M_PI*(i/(float)n);
      float c=std::cos(theta), s=std::sin(theta);
      glVertex3f(c, s, 0);
   }
   glEnd();
}

